#include <avr/io.h>
#include "tlc5916.h"
// LE pin macros
#define LE      PD7
#define LE_DDR  DDRD
#define LE_PORT PORTD
#define LE_HI   LE_PORT |= (1<<LE);
#define LE_LO   LE_PORT &= ~(1<<LE);
#define LATCH   LE_PORT ^= (1<<LE); LE_PORT ^= (1<<LE);
// OE pin macros
#define OE      PD6
#define OE_DDR  DDRD
#define OE_PORT PORTD
#define OE_HI   OE_PORT |= (1<<OE);
#define OE_LO   OE_PORT &= ~(1<<OE);
// SCK pin macros
#define SCK      PB0
#define SCK_DDR  DDRB
#define SCK_PORT PORTB
#define SCK_HI   SCK_PORT |= (1<<SCK);
#define SCK_LO   SCK_PORT &= ~(1<<SCK);
#define CLOCK    SCK_PORT ^= (1<<SCK); SCK_PORT ^= (1<<SCK);
// MOSI pin macros
#define MOSI      PD1
#define MOSI_DDR  DDRD
#define MOSI_PORT PORTD
#define MOSI_HI   MOSI_PORT |= (1<<MOSI);
#define MOSI_LO   MOSI_PORT &= ~(1<<MOSI);
// MISO pin macros
#define MISO      PD0
#define MISO_DDR  DDRD
#define MISO_PORT PORTD
#define MISO_HI   MISO_PORT |= (1<<MISO);
#define MISO_LO   MISO_PORT &= ~(1<<MISO);
// spi macros
#define SPI_DISABLE     //usart_disable();
#define SPI_ENABLE      //usart_enable();
#define SPI_TRANSMIT(x) softSPI(x); //usart_mspim_transceive(x);
// tlc5916 modes
#define NORMAL 0
#define SPECIAL 1
// function definitions
void tlc5916_init (void) {
  LE_DDR |= (1<<LE);
  OE_DDR |= (1<<OE);
  SCK_DDR |= (1<<SCK);
  MOSI_DDR |= (1<<MOSI);
  MISO_DDR |= (1<<MISO);
  OE_HI
  LE_LO
}
void mode (uint8_t newmode) {
  SPI_DISABLE
  SCK_LO
  LE_LO
  OE_HI
  CLOCK
  OE_LO
  CLOCK
  OE_HI
  CLOCK
  if (newmode == SPECIAL) {
    LE_HI
    CLOCK
    LE_LO
    CLOCK
  } else {
    CLOCK
    CLOCK
  }
  SPI_ENABLE
}
void softSPI (uint8_t data) {
  SCK_LO
  uint8_t i;
  for (i = 0; i < 8; i++) {
    if ((data & (1<<(7-i)))) {
      MOSI_HI
    } else {
      MOSI_LO
    }
    CLOCK
  }
}
void tlc5916_out (uint8_t data, uint8_t device) {
  static uint8_t out[TLC5916_QTY] = {0x00, 0x00};
  uint8_t i;
  out[(uint8_t)device] = data;
  for (i = 0; i < TLC5916_QTY; i++) {
    SPI_TRANSMIT (out[i])
  }
  LATCH
  OE_LO
}
void tlc5916_cur (uint8_t data, uint8_t device) {
  static uint8_t cur[2] = {0xFF, 0xFF};
  uint8_t i;
  uint8_t tmp = 0x00;
  for (i = 0; i < 8; i++)
    tmp |= ((data >> (7 - i))<<i);
  cur[device] = tmp;
  mode (SPECIAL);
  for (i = 0; i < 2; i++) {
    SPI_TRANSMIT (cur[i])
  }
  LATCH
  mode (NORMAL);
  OE_LO
}
